import numpy as np
import stan
import arviz as az

#
#    Simple Regression Model. See Gelman et al., 2014, p. 177
#
stan_code = """
    data {
        int N;
        array[N] real y;
        array[N] real x;      
    }
    parameters {
        real p_a;
        real p_b;
        real<lower=0.0> p_sgm;
    }
    transformed parameters {
        real a;
        real b;
        real sgm;
        a = p_a * 5.0;
        b = p_b * 1.0;
        sgm = p_sgm * 1.0 + 0.001;
    }
    model {
        p_a ~ normal(0.0, 10.0);
        p_b ~ normal(0.0, 10.0);
        p_sgm ~ exponential(0.1);
        for (i in 1:N) {
            y[i] ~ normal(a + b*x[i], sgm);
        }
    }
    generated quantities {
        array[N] real LL;         //  Log-Likelihood
        for (i in 1:N)
            LL[i] = normal_lpdf(y[i] | a + b*x[i], sgm);
    }
"""
#
#       Data. See Gelman et al., 2021.
#       from 15 ElectionsEconomy data (with the last record deleted)
#
xy = [[44.6, 2.4], [57.76, 2.89], [49.91, 0.85], [61.34, 4.21],
      [49.6, 3.02],[61.79, 3.62], [48.95, 1.08], [44.7, -0.39],
      [59.17, 3.86],[53.94, 2.27],[46.55, 0.38], [54.74, 1.04],
      [50.27, 2.36],[51.24, 1.72],[46.32, 0.1]]
x = np.array(xy).T[1]
y = np.array(xy).T[0]
Data = {'N':len(y), 'y':y, 'x':x} 
sm = stan.build(stan_code, data = Data)
fit = sm.sample(num_samples = 5000, num_warmup = 5000)

inference_data = az.from_pystan(posterior = fit, log_likelihood = {'LogL':'LL'},
                                posterior_model = sm)
smmry = az.summary(inference_data, var_names = ['a', 'b', 'sgm'])
print(smmry)

results_loo = az.loo(inference_data, scale='deviance')

print(f'\n LOO-CV = {results_loo.elpd_loo:.3f}') 
      
print('\nWAIC:')
results_waic = az.waic(inference_data, scale='deviance')

print(f'\n WAIC = {results_waic.elpd_waic:.3f}')

print('\n\nStatistics')
print('\nLOOCV:\n', results_loo)
print('\nWAIC:\n', results_waic)
